/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is SHA 180-1 Reference Implementation (Compact version) 
 * 
 * The Initial Developer of the Original Code is Paul Kocher of 
 * Cryptography Research. Portions created by Paul Kocher are 
 * Copyright (C) 1995-9 by Cryptography Research, Inc. All 
 * Rights Reserved. 
 * 
 * C++ Modifications created by Dave Smith for Jabber, Inc. are 
 * Copyright (c) 2001 Jabber, Inc. 
 * 
 * Contributor(s): 
 * 
 * Paul Kocher 
 * Dave Smith, Jabber, Inc.
 */

#include "bedrock.h"
using namespace bedrock;
using namespace std;


#define SHA_ROTL(X,n) (((X) << (n)) | ((X) >> (32-(n))))

SHAHasher::SHAHasher(void)
{
    reset();
}

/** 
 * Feed additional data into the hashing engine.
 * @throws exception::ResetRequired if toString() has been
 * called and reset() has not.
 */
void SHAHasher::hash(const char* data, int slen)
{
    if (_finalized)
        throw exception::ResetRequired();

    unsigned char* dataIn = (unsigned char*)data;
    int len = (slen==-1)?strlen(data):slen;

    for (int i = 0; i < len; i++) 
    {
        _W[_lenW / 4] <<= 8;
        _W[_lenW / 4] |= (unsigned long)dataIn[i];
        if ((++_lenW) % 64 == 0) {

            // shaHashBlock
            {
                int t;
                unsigned long A,B,C,D,E,TEMP;

                for (t = 16; t <= 79; t++)
                    _W[t] = 
                        SHA_ROTL(_W[t-3] ^ _W[t-8] ^ _W[t-14] ^ _W[t-16], 1);

                A = _H[0];
                B = _H[1];
                C = _H[2];
                D = _H[3];
                E = _H[4];

                for (t = 0; t <= 19; t++) {
                    TEMP = SHA_ROTL(A,5) + (((C^D)&B)^D)     + E + _W[t] 
                           + 0x5a827999L;
                    E = D; D = C; C = SHA_ROTL(B, 30); B = A; A = TEMP;
                }
                for (t = 20; t <= 39; t++) {
                    TEMP = SHA_ROTL(A,5) + (B^C^D)           + E + _W[t] 
                           + 0x6ed9eba1L;
                    E = D; D = C; C = SHA_ROTL(B, 30); B = A; A = TEMP;
                }
                for (t = 40; t <= 59; t++) {
                    TEMP = SHA_ROTL(A,5) + ((B&C)|(D&(B|C))) + E + _W[t] 
                           + 0x8f1bbcdcL;
                    E = D; D = C; C = SHA_ROTL(B, 30); B = A; A = TEMP;
                }
                for (t = 60; t <= 79; t++) {
                    TEMP = SHA_ROTL(A,5) + (B^C^D)           + E + _W[t] 
                           + 0xca62c1d6L;
                    E = D; D = C; C = SHA_ROTL(B, 30); B = A; A = TEMP;
                }

                _H[0] += A;
                _H[1] += B;
                _H[2] += C;
                _H[3] += D;
                _H[4] += E;
            }

            _lenW = 0;
        }
        _sizeLo += 8;
        _sizeHi += (_sizeLo < 8);
    }
}

/**
 * Finalize the data being hashed and return a hexidecimal
 * string representation of the hash. It is illegal to call 
 * SHAHasher::hash() again until after reset() is called.
 */
string SHAHasher::toString()
{
    char pad0x80 = 0x80;
    char pad0x00 = 0x00;
    char padlen[8];
    int bitshifts[4] = { 24, 16, 8, 0 };
    int i;

    if (!_finalized)
    {
        /* Pad with a binary 1 (e.g. 0x80), then zeroes, then length
         */
        padlen[0] = (unsigned char)((_sizeHi >> 24) & 255);
        padlen[1] = (unsigned char)((_sizeHi >> 16) & 255);
        padlen[2] = (unsigned char)((_sizeHi >>  8) & 255);
        padlen[3] = (unsigned char)((_sizeHi >>  0) & 255);
        padlen[4] = (unsigned char)((_sizeLo >> 24) & 255);
        padlen[5] = (unsigned char)((_sizeLo >> 16) & 255);
        padlen[6] = (unsigned char)((_sizeLo >>  8) & 255);
        padlen[7] = (unsigned char)((_sizeLo >>  0) & 255);
        hash(&pad0x80, 1);
        while (_lenW != 56)
            hash(&pad0x00, 1);
        hash(padlen, 8);

        _finalized = true;
    }

    /* Output hash
     */
    char  hashbuf[41];
    char* pos = hashbuf;
    for (i = 0; i < 20; i++) {
        snprintf(pos, 3, "%02x", 
            (unsigned char)(_H[i / 4] >> bitshifts[i % 4]));
        pos += 2;
    }

    return string(hashbuf);
}

/**
 * Reset the hashing engine so that a new hash can be
 * generated.
 */
void SHAHasher::reset()
{
    _lenW   = 0;
    _sizeHi = 0;
    _sizeLo = 0;

    /* Initialize H with the magic constants (see FIPS180 for constants)
     */
    _H[0] = 0x67452301L;
    _H[1] = 0xefcdab89L;
    _H[2] = 0x98badcfeL;
    _H[3] = 0x10325476L;
    _H[4] = 0xc3d2e1f0L;

    for (int i = 0; i < 80; i++)
        _W[i] = 0;

    _finalized = false;
}
