//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      ThreadPool.cpp
// Description:   bedrock::ThreadPool implementation
// Created at:    Thu Jul 19 10:00:51 2001
// Modified at:   Mon Jul 30 19:31:28 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
// $Id: ThreadPool.cpp,v 1.1 2001/12/16 20:02:26 dr_pepper Exp $
//============================================================================

#include "bedrock.h"
using namespace bedrock;
using namespace std;


class RefCountedSentinel
{
public:
    RefCountedSentinel(int initial_count, bedrock::callback::Functor* f)
	: _ref_count(initial_count), _callback(f)
	{}
    void releaseReference()
	{
	    _ref_lock.lock();
	    _ref_count--;
	    if (_ref_count == 0)
	    {
		(*_callback)();
		delete _callback;
		_ref_lock.unlock();
		delete this;
		return;
	    }
	    _ref_lock.unlock();
	}
private:
    Mutex              _ref_lock;
    int                _ref_count;
    callback::Functor* _callback;
};

/**
 * Default constructor
 * @param workers Number of threads this pool should use
 * @param active  Indicates whether worker threads are 
 * immediately created; defaults to TRUE
 */
ThreadPool::ThreadPool(unsigned int workers, bool active)
    : _current_worker(0)
{
    // Size up workers vector
    _workers.reserve(workers);

    // Initialize workers
    for (unsigned int i = 0; i < workers; i++)
        _workers.push_back(new Worker(active));
}

ThreadPool::~ThreadPool()
{
    // Cleanup; implicitly stops and joins all worker threads
    for_each(_workers.begin(), _workers.end(), Worker::_releaseWorker);
}

/**
 * Enqueue a callback for processing by any of the 
 * workers. No guarantee is made as to which worker
 * will process this callback 
 * @param cb Callback object for worker thread to process
 * @param delete_functor Should the thread pool free the functor after
 *                        it calls it?
 * @returns ID of worker thread used for processing
 */
int ThreadPool::enqueue(bedrock::callback::Functor* cb, bool delete_callback)
{
    int result = getNextThreadID();
    
    _workers[result]->enqueue(cb, delete_callback);
    
    return result;
}

/** 
 * Enqueue a callback for processing with a particular
 * worker. The key is used to select a worker; if the
 * same key is used multiple times, requests will be
 * processed in-order.
 * @param cb  Callback object for worker thread to process
 * @param id ID of worker thread to process with
 * @param delete_functor Should the thread pool free the functor after
 * it callse it?
 */
void ThreadPool::enqueue(bedrock::callback::Functor* cb, int id, 
                         bool delete_callback)
{
    _workers[id % _workers.size()]->enqueue(cb, delete_callback);
}

/**
 * Enqueue a special sentinel on all the threads in the thread
 * pool. When the sentinel has been processed by all threads, the
 * user-supplied functor will be executed. This provides a way to know
 * when all the threads in the pool have processed past a certain
 * point. Note that cb will be executed on the last thread to process
 * the sentinel. 
 * @param cb User-supplied functor to execute after all threads have
 * processed the sentinel.
 */
void ThreadPool::enqueueSentinel(bedrock::callback::Functor* cb)
{
    RefCountedSentinel* r = new RefCountedSentinel(_workers.size(), cb);
    for (vector<Worker*>::iterator it = _workers.begin(); it != _workers.end();
	 it++)
    {
	(*it)->enqueue(callback::wrap(r, &RefCountedSentinel::releaseReference), true);
    }
}

/**
 * Get the next worker thread id for enqueing into.
 * @returns ID of a worker thread.
 */
int ThreadPool::getNextThreadID(void)
{
    _current_worker_lock.lock();

    // Get the current worker ID, then increment; the modulus
    // ensures that we have proper wraparound for worker selection
    int result = (_current_worker++) % _workers.size();

    _current_worker_lock.unlock();

    return result;
}

/**
 * Pause worker threads. To restart them, call ThreadPool::resume
 */
void ThreadPool::pause()
{
    for_each(_workers.begin(), _workers.end(), Worker::_pauseWorker);
}

/**
 * Resume paused worker threads. 
 */
void ThreadPool::resume()
{
    for_each(_workers.begin(), _workers.end(), Worker::_resumeWorker);
}

// ---------------------
// ThreadPool::Worker
// ---------------------
ThreadPool::Worker::Worker(bool active)
    : _callbacks_cv(_lock), _active(active)
{
    pthread_create(&_id, NULL, ThreadPool::Worker::run, this);
}

void ThreadPool::Worker::_releaseCallback(FunctorHolder& fh)
{
    delete fh.fn;
}

ThreadPool::Worker::~Worker()
{
    pthread_cancel(_id);
    pthread_join(_id, NULL);

    _lock.lock();

    // Process remaining callbacks
    // Actually, don't process the callbacks... once the threadpool
    // is destroyed, all the other objects are almost sure to be.
    // We need an "about to delete worker" function if we want to
    // do a two stage deletion.
    for_each(_callbacks.begin(), _callbacks.end(),
             _releaseCallback);
             
    _lock.unlock();
}

void ThreadPool::Worker::enqueue(bedrock::callback::Functor* cb,
                                 bool delete_callback)
{
    _lock.lock();

    assert(cb != NULL);

    FunctorHolder fh = {cb, delete_callback};
    _callbacks.push_back(fh);

    if (_active)
        _callbacks_cv.signal();

    _lock.unlock();
}

void ThreadPool::Worker::pause()
{
    _lock.lock();

    _active = false;

    _lock.unlock();
}

void ThreadPool::Worker::resume()
{
    _lock.lock();

    _active = true;

    if (!_callbacks.empty())
        _callbacks_cv.signal();

    _lock.unlock();
}

// The stl containers are not at all thread safe, so you have to lock
// while you are manipulating them, even if you know that you are the
// only one pulling stuff out of them, and that you are only pulling
// stuff out that is already there.
void* ThreadPool::Worker::run(void* arg)
{
    ThreadPool::Worker *worker = (ThreadPool::Worker *)arg;

    worker->_lock.lock();

    while (1)
    {
        while (!worker->_callbacks.empty() && worker->_active)
        {
            FunctorHolder fh = worker->_callbacks.front();
            worker->_callbacks.pop_front();
            worker->_lock.unlock();
	    try
	    {
		(*(fh.fn))();
	    }
	    catch (...)
	    {
		cerr << "Programmer error. Uncaught exception within ThreadPool. Aborting..." << endl;
		exit(-1);
	    }
            if (fh.delete_flag)
                delete fh.fn;
            pthread_testcancel();
            worker->_lock.lock();
        }
        worker->_callbacks_cv.wait();
    }

    return NULL;
}
