//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      Timer.cpp
// Description:   bedrock::Timer implementation
// Created at:    Thu Jul 19 10:00:51 2001
// Modified at:   Tue Jul 31 15:20:36 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
// $Id: Timer.cpp,v 1.1 2001/12/16 20:02:26 dr_pepper Exp $
//============================================================================

#include "bedrock.h"
using namespace bedrock;
using namespace std;


class P_DelEntryEq 
{
        int _key;
    public:
        P_DelEntryEq(int key) : _key(key) {}
        bool operator()(const Timer::Entry* e) const
        { 
            if (e->key == _key)
            {
                delete e->fn;
                delete e;
                return true;
            }
            else
                return false;
        }
};

class P_EntryComplete
{
    int _counter;
public:
    P_EntryComplete(int counter) : _counter(counter) {}
    bool operator()(Timer::Entry* e)
        {
            if (_counter % e->timelapse == 0)
            {
                if (e->runcount > 0)
                    e->runcount--;
                if (e->runcount == 0)
                {
                    // Enqueue and delete afterwards
                    e->tpool->enqueue(e->fn, e->tpoolkey);
                    delete e;
                    return true;
                }
                else
                {
                    // Enqueue and save afterwards
                    e->tpool->enqueue(e->fn, e->tpoolkey, false);
                    return false;
                }
            }
            return false;
        }
};

template<class List, class Predicate>
void apply_remove_if(List& l, Predicate p)
{
    typename List::iterator first = l.begin();
    typename List::iterator last  = l.end();
    while (first != last) {
        typename List::iterator next = first;
        ++next;
        if (p(*first)) 
            l.erase(first);
        first = next;
    }
}

inline
void deleteEntry(Timer::Entry* e)
{
    delete e;
}

Timer::Timer(int interval)
    : _next_key(1), _interval(interval), _counter(1)
{
    pthread_create(&_threadid, NULL, &Timer::run, this);
}

Timer::~Timer()
{
    pthread_cancel(_threadid);
    pthread_join(_threadid, NULL);

    // Walk the list and delete all the entries
    for_each(_entrylist.begin(), _entrylist.end(), deleteEntry);
}

/**
 *  Schedule an event with this timer. 
 *  @param timelapse The delay before firing this event, or between event
 *                   occurences.
 *  @param p         The threadpool in which to run this event.
 *  @param threadkey The threadpool key to use for this event.
 *  @param fn        The event to fire.
 *  @param runcount  The number of times to execute this event.  Supply a value
 *                   of -1 (the default) for infinite occurances.
 *  @returns A key associated with this event that can be used to unschedule
 *           the event at a later time.
 */
int Timer::schedule(int timelapse, ThreadPool* p, int threadkey,
                     callback::Functor* fn, int runcount)
{
    Entry* entry = new Entry; 
    
    entry->runcount  = runcount;
    entry->timelapse = timelapse;
    entry->tpool     = p;
    entry->tpoolkey  = threadkey;
    entry->fn        = fn;

    MutexProxy mp(_lock);

    entry->key = ++_next_key;
    _entrylist.push_back(entry);

    return entry->key;
}

/**
 *  Unschedule an event that was previously scheduled with this timer.
 *  @param key The key associated with the event to cancel.  This is the value
 *             returned by schedule when the event was initially scheduled.
 */
void Timer::unschedule(int key)
{
    MutexProxy mp(_lock);
    
    apply_remove_if(_entrylist, P_DelEntryEq(key));
}

/**
 * This function gets the current value of the counter.  Basically,
 * the number of _interval seconds that have gone by since the Timer
 * was created.  This is useful for stamping an object's last access
 * time and then timing it out if that stamp is a certain age.  I put
 * this in here because locking a mutex, getting a value, and then
 * unlocking the mutex is something on the order of 12 times faster
 * than calling time(), at least on Linux.
 */
unsigned int Timer::getCounter(void) 
{
    MutexProxy mp(_lock);
    return _counter;
}

void* Timer::run(void* arg)
{
    Timer* timer = (Timer *)arg;
    timer->run();

    return NULL;
}

void Timer::run(void)
{
    while (1)
    {
        sleep(_interval);
        pthread_testcancel();
        {
            MutexProxy mp(_lock);

            // The STL list<> template has a member function iterates across
            // all items in the list, and applies a predicate to each in order
            // to determine if the item should be removed from the list. We use
            // this method to walk all items in our timer list and remove them if
            // they have "expired"
            apply_remove_if(_entrylist, P_EntryComplete(_counter));
            _counter++;
        }
    }
}




