//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      bedrock.h
// Description:   Primary header file
// Created at:    Thu Jul 19 14:51:40 2001
// Modified at:   Wed Aug  1 22:21:54 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
// $Id: bedrock.h,v 1.1 2001/12/16 20:02:26 dr_pepper Exp $
//============================================================================
#ifndef _H_BEDROCK
#define _H_BEDROCK

#include <unistd.h>
#include <math.h>
#include <errno.h>
#include <pthread.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/poll.h>
#include <assert.h>

#include <list>
#include <queue>
#include <vector>
#include <string>
#include <strstream>
#include <set>
#include <algorithm>
#include <map>

#ifdef TEST
#define TESTER(s) friend class s;
#else
#define TESTER(s)
#endif

#include "bedrock_callbacks.h"
#include "bedrock_debug.h"

/**
   Foundational classes for Jabber components.
*/
namespace bedrock
{
    // Forward declarations
    class ThreadPool;

    /**
     * Wrapper object for pthread mutexes.
     */
    class Mutex
    {
    public:
        Mutex() 
            { abortingCall(pthread_mutex_init(&_lock, NULL)); }
        ~Mutex()
            { abortingCall(pthread_mutex_destroy(&_lock)); }

        void lock()
            { abortingCall(pthread_mutex_lock(&_lock)); }
        void unlock()
            { abortingCall(pthread_mutex_unlock(&_lock)); }

        pthread_mutex_t* getNative()
            { return &_lock; }

    private:
        TESTER(MutexTest)

        pthread_mutex_t _lock;
    };

    /**
     * Proxy object for ensuring that a Mutex gets unlocked
     * when the stack unwinds.
     */
    class MutexProxy
    {
    public:
        MutexProxy(Mutex& m)
            : _m(m)
            { _m.lock(); }
        ~MutexProxy()
            { _m.unlock(); }
    private:
        Mutex& _m;
    };

    /**
     * Wrapper object for pthread condition variables
     */
    class CondVariable
    {
    public:
        CondVariable(Mutex& m);
        ~CondVariable();

        void signal();
        void signalAll();
        void wait();
    protected:
        static void cancellation_cleanup(void* arg);
    private:
        TESTER(CondVariableTest)
        Mutex&           _mutex;
        pthread_cond_t   _condition;
    };


    /**
     * Provides an easy way to execute assorted functions 
     * in the context of a thread pool.
     */
    class ThreadPool
    {
    public:
        ThreadPool(unsigned int workers, bool active = true);
        ~ThreadPool();
    
        int  enqueue(bedrock::callback::Functor* cb, bool 
		     delete_functor = true);
        void enqueue(bedrock::callback::Functor* cb, int id, 
                     bool delete_functor = true);
	void enqueueSentinel(bedrock::callback::Functor* cb);
        int  getNextThreadID(void);
        void pause();
        void resume();
    private:
        TESTER(ThreadPoolTest)

        class Worker
        {
        public:
            Worker(bool active = true);
            ~Worker();
    
            void enqueue(bedrock::callback::Functor* cb, bool delete_callback);
            void pause();
            void resume();
        protected:
            bedrock::callback::Functor* dequeue();
            static void*                run(void* arg);

        private:
            struct FunctorHolder
            {
                callback::Functor* fn;
                bool               delete_flag;
            };

            static void _releaseCallback(FunctorHolder& fh);

            TESTER(ThreadPoolTest)
            friend class ThreadPool;
            Mutex                  _lock;
            std::list<FunctorHolder>    _callbacks;
            CondVariable           _callbacks_cv;
            pthread_t              _id;
            bool                   _active;

        static void _releaseWorker(Worker* w)
            { delete w; }
        static void _pauseWorker(Worker* w)
            { w->pause(); }
        static void _resumeWorker(Worker* w)
            { w->resume(); }
        static void _executeWorkerCallback(bedrock::callback::Functor* cb)
            { (*cb)();  delete cb; }
        };

        std::vector<Worker*> _workers;
        unsigned int    _current_worker;
        Mutex           _current_worker_lock;
    };

    class Timer
    {
    public:
        Timer(int interval);
        ~Timer();

        int          schedule(int timelapse, ThreadPool* p, int threadkey, 
			      callback::Functor* fn, int runcount = -1);
        void         unschedule(int key);
        unsigned int getCounter(void);

        struct Entry
        {
            int                runcount;
            int                timelapse;
            ThreadPool*        tpool;
            int                tpoolkey;
            callback::Functor* fn;
            int                key;
        };

    protected:
        static void* run(void* arg);
        void run();
    private:
        pthread_t     _threadid;
        Mutex         _lock;
        std::list<Entry*>  _entrylist;
        int           _next_key; 
        int           _interval;
        unsigned int  _counter;
    };

    /**
     * Retrieve a timestamp in ISO 8601 form.
     */    
    std::string getTimeStamp();
    std::string getTimeStamp(time_t time);

    /**
       Generates SHA-1 hashes.
    */
    class SHAHasher {
    public:    
        /**
           Default constructor 
        */
        SHAHasher(void);

        void hash(const char* data, int len = -1);
        void hash(const std::string& data)
            {  hash(data.c_str(), data.size()); }

        std::string toString();
        void   reset();

        struct exception
        {
            class ResetRequired {};
        };

    private:    
        TESTER(SHAHasherTest)

        unsigned long _H[5];
        unsigned long _W[80];
        int           _lenW;
        unsigned long _sizeHi; 
        unsigned long _sizeLo;

        bool          _finalized;
    };

    /** 
	Application loop singleton. This object provides proper signal
	handling for threaded programs and a way for controller
	objects to shutdown the application in a graceful way.
    */
    class Application 
    {
    public:
	/**
	   Put the current thread to sleep until it is time to exit
	   the current application.
	   @see Application::stop
	*/
	static int start();

	/**
	   Kick the main thread and tell it to wakeup and complete any
	   processing necessary for shutting down. You <b>must</b>
	   call Application::start before calling this method. If you
	   do not, this function will <b>ASSERT</b>.
	   @param rc Code to return from Application::start
	   @param message Optional message describing why this
	   shutdown is occurring
	*/
	static void stop(int rc, const std::string& message = "");
	
	/**
	 * Exit immediately
	 * @param ec error code
	 * @param message Textual message describing why this shutdown is 
	 *                occurring
	 */
	static void exit(int ec, const std::string& message = "");


	/** 
	    Register a OS signal which should fire an Application::stop
	    event. Do not call this function after calling Application::start.
	*/
	static void registerShutdownSignal(int signal);
	
    private:
	// Hide all constructors so no one will
	// try and instantiate an Application object
	Application()
	    : _shutdown_signal(_shutdown_lock), _shutdown_code(0), 
	      _signal_occurred(false), _started(false), _shutting_down(false)
	    {}
	~Application();
	Application(const Application& a);
	Application& operator=(const Application& a);

	// Signal handler
	static void signal_callback(int signal_id);

	// Static singleton instance
	static Application* _app_instance;

	// Private members
	Mutex        _shutdown_lock;
	CondVariable _shutdown_signal;
	int          _shutdown_code;
	std::string       _shutdown_msg;
	bool         _signal_occurred;
	bool         _started;
	bool         _shutting_down;
    };
    
    /**
       Network related classes and functions.
    */
    namespace net
    {
        // Forward declarations
        class Socket; 
	class SocketEventListener; 
        class SocketWatcher; 
        class Resolver; 
        class Address; 
	class Karma;
	class KarmaFactory;

        /**
	 * Container for network addresses.
	 */
        class Address
        {
        public:
	    Address(const std::string& hostname, const std::string& service, 
		    unsigned int port=0)
		: _hostname(hostname), _service(service), _default_port(port)
		{}

	    Address() : _default_port(0){}

            /**
	     * Retrieve the default port of this address	
	     */
            unsigned int getPort() const 
		{
		    if (_entries.empty())
			return _default_port;
		    else
			return _entries.begin()->second.port;
		}

            /**
	     * Retrieve the default resolved IP of this address. If the
	     * address hasn't been resolved, this will return an empty
	     * string
	     */
            std::string getIP() const 
                { 
		    if (_entries.empty())
			return "";
		    else
			return _entries.begin()->second.ip;
		}

            /**
	     * Retrieve the hostname of this address.
	     */
            const std::string& getHostname() const
                { return _hostname; }

            /**
               Retrieve the service of this address. Specifically,
               this is used for SRV lookups 
            */
            const std::string& getService() const
                { return _service; }

            /**
               Mutator/setter for this hostname
            */
            void setHostname(const std::string& hostname)
                { _hostname = hostname; }
            /**
               Mutator/setter for this service
            */
            void setService(const std::string& service)
                { _service = service; }

	    void setDefaultPort(unsigned int port)
		{ _default_port = port; }

	    /**
	       Add a IP/port entry to this Address
	    */
	    void addEntry(int priority, const std::string& ip, unsigned int port = 0)
		{ 
		    if (port == 0)
		    {
			_entries.insert(std::make_pair(priority, Entry(priority, ip, _default_port))); 
		    }
		    else
		    {
			_entries.insert(std::make_pair(priority, Entry(priority, ip, port))); 
		    }
		}

            static void   inet_ntoa(unsigned char* addr, std::string& result);
	    static std::string inet_ntoa(unsigned char* addr);

	    struct Entry
	    {
		int    priority;
		std::string ip;
		int    port;
		Entry(int apriority, const std::string& aip, int aport)
		    : priority(apriority), ip(aip), port(aport)
		    {}
	    };

	    typedef std::map<int, Entry>::iterator iterator;
	    typedef std::map<int, Entry>::const_iterator const_iterator;

	    iterator begin()
		{ return _entries.begin(); }
	    const_iterator begin() const
		{ return _entries.begin(); }
	    iterator end()
		{ return _entries.end(); }
	    const_iterator end() const
		{ return _entries.end(); }

	    unsigned int size()
		{ return _entries.size(); }
	    bool empty()
		{ return _entries.empty(); }
	    void clear()
		{ _entries.clear(); }

        private:
            friend class Resolver;

	    std::map<int, Entry> _entries;
	    std::string          _hostname;
	    std::string          _service;
	    unsigned int    _default_port;
        };

	class KarmaFactory
	{
	public:
	    KarmaFactory(ThreadPool& pool);
	    Karma *create(int read_bytes, int write_bytes, ThreadPool& tpool,
			  int tkey);
	private:

	    friend class Karma;
	    
	    void _register(Karma* karma);
	    void _unregister(Karma* karma);
	    void _resetKarma();
	    static void _do_reset(Karma* karma);

	    TESTER(KarmaTest)
	    std::list<Karma *> _k_list;
	    Mutex         _lock;
	    Timer         _timer;
	};

	/**
	 * Tracks I/O -- knows when a socket has done too much
	 */
	class Karma
	{
	public:

	    enum KarmaType
	    {
		Read,
		Write
	    };

	    Karma(const Karma&);
	    Karma(const Karma&, ThreadPool& tpool, int threadkey);
	    ~Karma();

	    bool mayReadMore(int bytes = 0);
	    bool mayWriteMore(int bytes = 0);
	    void callWhenReset(bedrock::callback::Functor* cb, KarmaType kt);
	private:
	    friend class KarmaFactory;
	    Karma();
	    Karma(int read_bytes, int write_bytes, ThreadPool &tpool, int tkey,
		  KarmaFactory& factory);

	    void _reset();

	    TESTER(KarmaTest)
	    int                              _read_bytes;
	    int                              _read_bytes_left;
	    bedrock::callback::Functor*      _read_cb;

	    int                              _write_bytes;
	    int                              _write_bytes_left;
	    bedrock::callback::Functor*      _write_cb;

	    Mutex                            _lock;

	    ThreadPool&                      _threadpool;
	    int                              _threadkey;

	    KarmaFactory&                    _factory;

	};

	/**
	 * Limits the amount of incoming connections by IP
	 */
	class ConnectionLimit
	{
	public:
	    ConnectionLimit(ThreadPool &tpool, int connections_per_second);
	    void addBlacklistItem(const char *dotted_ip);
	    bool allowConnection(unsigned int ip);
	    
	private:
	    void _resetConnectionList(void);
	    
	    // Banned addresses
	    std::set<unsigned int>       _blacklist; 
            // Connections remaining this second -- keyed by inet address.
	    std::map<unsigned int, int>  _connections; 
	    const int               _connects_per_second;
	    Timer                   _timer;
	    Mutex                   _lock;
	};
	
        /**
           Provides DNS/SRV address resolution. 
        */
        class Resolver
	{
        public:
            Resolver(int resolverthreads = 2);
            ~Resolver();

            void        lookup(Address* result, 
			       bedrock::callback::Functor* cb, 
			       ThreadPool* tp, int tkey);
            static void lookupSRV(Address* result);
            static void lookupDNS(Address* result);

        private:
            Resolver(const Resolver& r);
            TESTER(ResolverTest)
            static void executeLookup(Address* result, 
                                      bedrock::callback::Functor* cb,
                                      ThreadPool* tp, int tkey);
            ThreadPool _resolverpool;
        };

        /**
           Monitors multiple sockets for events.
        */
        class SocketWatcher
        {
        private:
            enum Event { 
		NOOP,
		Read,
		Write,
		ReadWrite,
		Closing,
		Closed
            };

            struct SocketInfo
            {
		int                  id;
                Socket*              socket;
                SocketWatcher::Event event;
            };
	    typedef std::list<SocketInfo>::iterator SocketInfoHandle;

        public:
            SocketWatcher(ThreadPool& pool, unsigned int maxsockets);
             ~SocketWatcher();

            Socket* createListenSocket(SocketEventListener* listener,
                                       const char* ip, 
                                       int port,
                                       bool reuseaddr = false,
                                       int thread_key = -1,
				       Karma* karma = 0,
				       ConnectionLimit* cl = 0);
            Socket* createConnectSocket(SocketEventListener* listener,
                                        const char* hostname, 
                                        const char* service, 
                                        int defaultport,
                                        int thread_key = -1,
					Karma* karma = 0);

            /**
             *  Get this SocketWatcher's resolver.
             */
            Resolver& getResolver()
                { return _resolver; }


            struct exception
            {
                class WatcherFull {};
            };

	    // Socket interface functions
            SocketInfoHandle acquireHandle(Socket* s);
            void             watchSocket(const SocketInfoHandle& h, SocketWatcher::Event event);
        protected:
	    // Internal socket management functions
	    void             releaseHandle(SocketInfoHandle& handle);
            void             cleanupSocket(SocketInfoHandle h);

	    // Watcher thread functions
            static void* beginProcessing(void* arg);
            void process(void);
            void processEvents(void);
            void processPending(void);

            static void _closeSocket(SocketInfo&);
        private:
            SocketWatcher(const SocketWatcher& w);
	    SocketWatcher& operator=(const SocketWatcher& w);

            enum State
            {
                Running,
                Shutdown,
                Stopped
            };

            TESTER(SocketWatcherTest)

            friend class Socket;

            // Constant members (i.e. members which never change during
            // the lifetime of the object)
            Resolver             _resolver;         // Internal resolver
            unsigned int         _socket_cnt;       // Max # of sockets
            int                  _poll_wakeup_pipe;
            ThreadPool&          _threadpool;
            pthread_t            _watcher_thread;   // Processing thread ID
	    Mutex                _lock;             // Lock to protect 
                                                    // non-constant members

            // Non-constant members (protected by _lock)
            std::set<int>                 _available_slots;  // List of available slots
	    std::vector<pollfd>           _fds;            
            std::list<int>                _pending;          // Socket slots w/ changes pending
	    std::list<SocketInfo>         _sockets;          // List of all the sockets we know about, watched or not
	    std::list<SocketInfoHandle>   _unwatched_sockets; 
	    std::vector<SocketInfoHandle> _watched_sockets;
	    int                      _close_pending_count; // Number of sockets in process of closing
            State                _state;            // State of the SocketWatcher system
            CondVariable         _stopped_cv;       // CV for notification when all sockets are gone
	    
	    
        };


        /**
           Basic network connectivity class. Represents a single
           network connection. Note that this class is instantiated by
           a SocketWatcher; it is not possible to instantiate a Socket
           without a watcher.
           @see SocketWatcher::createConnectSocket
           @see SocketWatcher::createListenSocket
        */
        class Socket
        {
        public:
            /**
	     * Retrieve the threadkey used by this socket.
	     * @returns The threadkey associated with this socket.
	     */
            int getThreadKey(void) { return _threadkey; }
         
            /**
             *  Set the SocketEventListener that this socket
             *  will inform when events take place.
             */
            void  setListener(SocketEventListener* l)
                { _listener = l; }

            /**
             *  Get the SocketEventListener that this socket
             *  will inform when events take place.
             */
            SocketEventListener* getListener() const
                { return _listener; }

            /**
             *  Get a copy of the address to which this socket is connected
             *  or is connecting.
             */
            Address getAddress() const
                { return _address; }

            /**
             *  Get the socket's underlying file descriptor.
             */
            int getFD() const
                { return _socket; }

	    void close();
            void resumeRead();
            void resumeAccept();
            void write(const char* data);

            /**
             *  Write data on the socket.  If the socket is not yet connected
             *  it will queue it for writing when the socket connection is
             *  completed.  Do not call this from an accepting socket or you
             *  will get an exception.
             */
            void write(const std::string& data)
                { write(data.c_str()); }

            struct exception
            {
                enum ErrorCode { InvalidHostname,
                                 SocketConnectError,
                                 InvalidSocketAccept,
                                 SocketWatcherFull,
                                 SocketAcceptError,
                                 SocketRecvError,
                                 SocketSendError,
                                 SocketErrorEvent
                };

		class BaseException {};

                template<int id>
                class ExceptionTemplate
		    : public BaseException
                    {
                    public:
                        ExceptionTemplate(int e = 0, const char* s = NULL)
                            : en(e), es(s) {}
                        int en;
                        const char* es;
                    };
                typedef ExceptionTemplate<0> BindError;
                typedef ExceptionTemplate<1> ListenError;
                typedef ExceptionTemplate<2> ErrorCreatingSocket;
            };
 
       protected:

            // Socket state
            enum State { Created, 
                         Listening,
                         Connecting,
                         Connected,
                         Closed,
                         Error
            };

            // Constructors and Destructors:
            Socket(SocketWatcher& watcher, ThreadPool& pool, 
                   int key, int socketfd = -1, Karma *karma = 0,
		   ConnectionLimit *cl = 0);
            ~Socket();

            void connect(SocketWatcher::SocketInfoHandle handle, const char* hostname, 
			 const char* service, int defaultport);
            void accept(SocketWatcher::SocketInfoHandle, const char* ip, unsigned int port, 
			bool reuseaddr = false);

            void processEvents(int revents)
                { _threadpool.enqueue(
                    callback::wrap(this, &Socket::onSocketEvent, revents), 
                    _threadkey);
                }

            void processClose(void)
		{    
		    _threadpool.enqueue(callback::wrap(this, 
						       &Socket::executeClose), 
					_threadkey);
		}

	    

            // Request writes/reads/close from the SocketWatcher
            void requestWrite(char *data);
            void requestRead(void);
            void requestAccept(void);


            // Protected callbacks
            void executeConnect(void);
            void executeAccept(void);
            void executeRead(void);
            void executeWrite(void);
            void executeError(exception::ErrorCode ec, int err);
            void executeClose(void);

            void onSocketEvent(short int revents);
            void onConnectResolved(void);

        private:
            Socket(const Socket& s);
	    Socket& operator=(const Socket& s);

            TESTER(SocketTest)
            TESTER(SocketWatcherTest)

            // This class considers SocketWatcher a friend so
            // that SocketWatcher can call onSocketEvent.
            friend class SocketWatcher;

            // Member variables:
            ThreadPool&            _threadpool;
            int                    _threadkey;
            int                    _socket;
            Address                _address;
            SocketEventListener*   _listener;
            std::list<char*>            _writequeue;
            char*                  _current_write_buffer;
            int                    _current_write_buffer_len;
            Socket::State          _state;
            bool                   _resolving; 
	    Karma*                 _karma;
	    ConnectionLimit*       _conn_limit;

	    // Watcher/owner information

            SocketWatcher&                  _watcher;
	    SocketWatcher::SocketInfoHandle _watcher_info;
        };


        /**
           Interface class for Socket events. Any object which
           implements these interfaces is eligible to recieve Socket
           events.
        */
        class SocketEventListener
        {
        public:
            virtual ~SocketEventListener() {}
            virtual bool onAccept(Socket* newsocket) = 0;
            virtual void onConnected(Socket *sock) = 0;
	    virtual void onInitialize(Socket *sock) = 0;
            virtual void onClose(Socket *sock) = 0;
            virtual void onError(Socket *sock, Socket::exception::ErrorCode ec, int err) = 0;
            virtual bool onRead(Socket *sock, const char* buf, int bufsz) = 0;
        };


    }

}

#endif
