//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      ConnectionLimit.cpp
// Description:   bedrock::net::ConnectionLimit -- implementation
// Created at:    Sat Jul 28 10:00:51 2001
// Modified at:   Mon Jul 30 19:28:03 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or execcutable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
//============================================================================
#include "bedrock.h"
using namespace bedrock;
using namespace bedrock::net;
using namespace std;

/**
 * Constructor
 * @param tpool Threadpool to execute clear event in
 * @param connections_per_second the number of connections permitted per IP
 *                               per second.  -1 allows infinite connections
 */
ConnectionLimit::ConnectionLimit(ThreadPool &tpool, int connections_per_second)
    :_connects_per_second(connections_per_second),  _timer(1)
{
    if (connections_per_second != -1)
    {
	_timer.schedule(1, &tpool, 0, callback::wrap(this, &ConnectionLimit::_resetConnectionList), -1);
    }
}

/**
 * Ads an IP to the set of blacklisted IPs
 * @param dotted_ip the ip to blacklist
 */
void ConnectionLimit::addBlacklistItem(const char *dotted_ip)
{
    unsigned int int_ip = inet_addr(dotted_ip);
    _lock.lock();
    _blacklist.insert(int_ip);
    _lock.unlock();
}

/**
 * Checks if it is permissible to allow a connection from a given IP
 * @param ip the ip to check
 * @return true iff it is acceptable to accept this connection.
 */
bool ConnectionLimit::allowConnection(unsigned int ip)
{
    MutexProxy mp(_lock);

    // Are they blacklisted?
    set<unsigned int>::iterator blacklist_it = _blacklist.find(ip);
    if (blacklist_it != _blacklist.end()){
	return false;
    }
    
    // Have we seen them too often recently?

    // First verify we're even monitoring that
    if (_connects_per_second == -1)
    { //we're not
	return true;
    }
    map<unsigned int,int>::iterator connections_it = _connections.find(ip);
    if ( connections_it != _connections.end() )
    {// They're in our list --  see if they've got connections remaining
	if (connections_it->second > 0)
	{
	    connections_it->second--;
	    return true;
	}
	else
	{
	    return false;
	}
    }
    else
    {	// Add them with one less than allowed since we assume we just
	// accepted this connection
	_connections.insert(make_pair(ip, _connects_per_second - 1));
    }
    return true;
}
	    
/**
 * Reset number of permissible connections per IP for non-blacklisted IPs.
 */
void ConnectionLimit::_resetConnectionList(void)
{
    MutexProxy mp(_lock);
    _connections.clear();
}

