//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      Karma.cpp
// Description:   bedrock::net::Karma -- implementation
// Created at:    Thu Jul 26 10:00:51 2001
// Modified at:   Mon Jul 30 19:28:20 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
//============================================================================
#include "bedrock.h"
using namespace bedrock;
using namespace bedrock::net;
using namespace std;


/**
 * @param read_bytes the number of bytes one is allowed to read per second
 *                   -1 means unlimited reads
 * @param write_bytes the number of bytes one is allowed to write per second
 *                   -1 means unlimited writes
 * @param tpool the threadpool to execute callbacks in (during reset)
 * @param tkey the thread key to use when enqueuing callbacks on reset
 * @param factory the factory creating us
 * @param factory_info 
 */
Karma::Karma(int read_bytes, int write_bytes, ThreadPool &tpool, int tkey,
	     KarmaFactory &factory) :
    _read_bytes(read_bytes), _read_bytes_left(read_bytes), _read_cb(0),
    _write_bytes(write_bytes), _write_bytes_left(write_bytes), _write_cb(0),
    _threadpool(tpool), _threadkey(tkey), _factory(factory) 
{}

Karma::Karma(const Karma& k) :
    _read_bytes(k._read_bytes), _read_bytes_left(_read_bytes), _read_cb(0),
    _write_bytes(k._write_bytes), _write_bytes_left(_write_bytes), _write_cb(0),
    _threadpool(k._threadpool), _threadkey(k._threadkey), _factory(k._factory)
{
    _factory._register(this);
}

Karma::Karma(const Karma& k, ThreadPool& tpool, int threadkey) :
    _read_bytes(k._read_bytes), _read_bytes_left(_read_bytes), _read_cb(0),
    _write_bytes(k._write_bytes), _write_bytes_left(_write_bytes), _write_cb(0),
    _threadpool(tpool), _threadkey(threadkey), _factory(k._factory)
{
    _factory._register(this);
}
    
/**
 * Unregisters from the factory
 */
Karma::~Karma()
{
    _factory._unregister(this);
}

/**
 * Updates number of bytes left for writing and informs caller
 * whether more reading is allowed at this time.
 * @param bytes the number of bytes just read
 * @return true iff the caller may continue reading
 */
bool Karma::mayReadMore(int bytes)
{
    bool return_val;
    if (_read_bytes == -1)
    {
	return_val = true;
    }
    else
    {
	_read_bytes_left -= bytes;
	return_val = (_read_bytes_left > 0);
    }
    return return_val;
}

/**
 * Updates number of bytes left for writing and informs caller
 * whether more writing is allowed at this time.
 * @param bytes the number of bytes just written
 * @return true iff the caller may continue writing
 */
bool Karma::mayWriteMore(int bytes)
{
    bool return_val;
    if (_write_bytes == -1)
    {
	return_val = true;
    }
    else
    {
	_write_bytes_left -= bytes;
	return_val = (_write_bytes_left > 0);
    }

    return return_val;
}

/**
 *  Registers a callback for when karma is reset
 *  @param cb the Functor to be called when karma is reset
 *  @param kt whether this should be called when read is reset or when write
 *            is reset
 */
void Karma::callWhenReset(bedrock::callback::Functor* cb, Karma::KarmaType kt)
{
    MutexProxy mp(_lock);
    if (kt == Karma::Read && _read_cb == NULL)
    {
	_read_cb = cb;
    }
    else if (kt == Karma::Write && _write_cb == NULL)
    {
	_write_cb = cb;
    }
    else
    {
	delete cb;
    }
}

/**
 *  Updates the amount one may read and write
 */
void Karma::_reset()
{
    MutexProxy mp(_lock);

    // Grant MOJO
    if (_read_bytes_left > 0)
	_read_bytes_left = _read_bytes;
    else
	_read_bytes_left += _read_bytes;

    if (_write_bytes_left > 0)
	_write_bytes_left = _write_bytes;
    else
	_write_bytes_left += _write_bytes;
    
    // Call 'em back if appropriate
    if (_read_cb && (_read_bytes_left > 0))
    {
	_threadpool.enqueue(_read_cb, _threadkey);
	_read_cb = NULL;
    }
    if (_write_cb && (_write_bytes_left > 0))
    {
	_threadpool.enqueue(_write_cb, _threadkey);
	_write_cb = NULL;
    }
}





