//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      Resolver.cpp
// Description:   bedrock::net::Resolver implementation
// Created at:    Thu Jul 19 10:00:51 2001
// Modified at:   Fri Jul 27 16:27:55 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
// $Id: Resolver.cpp,v 1.1 2001/12/16 20:02:26 dr_pepper Exp $
//============================================================================
#include "bedrock.h"
using namespace bedrock;
using namespace bedrock::net;
using namespace std;

Mutex* G_lookup_lock = new Mutex();

/**
 * Base constructor
 * @param resolverthreads Number of threads to create for
 * resolving addresses
 */
Resolver::Resolver(int resolverthreads)
                : _resolverpool(resolverthreads, true)
{
}

Resolver::~Resolver()
{
}

/**
 * Enqueue a DNS/SRV lookup.
 * @param result Reference to an Address struct which
 * has the initial DNS name and default port inside.
 * @param cb Pointer to a deferred callback which will be
 * invoked when the resolution is complete.
 * @param tp Pointer to the threadpool in which to execute the
 * provided callback
 * @param tkey Selector for threadpool thread to use for
 *             executing the callback
 */
void Resolver::lookup(Address* result, callback::Functor* cb, 
              ThreadPool* tp, int tkey)
{
    // Create a DeferredCallback that will ensapulate this result and callback
    callback::Functor *mycb = callback::wrap(Resolver::executeLookup, result, cb, tp, tkey);
    _resolverpool.enqueue(mycb);
}


void Resolver::executeLookup(Address* result, 
                 callback::Functor* cb,
                 ThreadPool* tp, int tkey)
{
    Resolver::lookupSRV(result);
    if (result->empty())
	Resolver::lookupDNS(result);

    // Queue up the callback
    tp->enqueue(cb, tkey);
}

/**
 * Perform a DNS lookup.           
 * @param result Reference to an Address struct which has the initial DNS 
 * name. The IP resulting from the DNS lookup will be stored in the Address 
 * and is accessible via Address::getIP.
 */
void Resolver::lookupDNS(Address *result)
{
    MutexProxy mp(*G_lookup_lock);
    hostent* hp = gethostbyname(result->getHostname().c_str());
    if (hp)
    {
	// Walk the list of resulting ips and store them in the 
	// result address
	for (int i = 0; i != hp->h_length / 4; i++)
	{
	    result->addEntry(i, Address::inet_ntoa((unsigned char*)hp->h_addr_list[i]));
	}
    }
}

#ifndef T_SRV
#define T_SRV 33
#endif

struct SRVRec
{
    int          priority;
    unsigned int port;
    string       host;
};

void Resolver::lookupSRV(Address *result)
{
    MutexProxy mp(*G_lookup_lock);
    // Execute a SRV query against the requested domain; according to
    // RFC 2782 there is a practical limit of 512 bytes on DNS
    // requests. As such, we'll use a buffer of 1024 to hold the DNS
    // response. The format of the result is specified in RFC 1035.
    unsigned char dns_result[1024];
    unsigned int dns_result_sz = res_querydomain(result->getService().c_str(),
						 result->getHostname().c_str(),
						 C_IN, T_SRV, 
						 dns_result,
						 1024);

    // Setup a pointer to the result header
    HEADER* result_header = (HEADER*)dns_result;

    // Ensure the SRV response meets with conditions specified in RFC 2782
    // 1.) reply has some data available
    // 2.) no error has occurred
    // 3.) there are 1 or more answers available
    if ((dns_result_sz < 0) 
	|| (ntohs(result_header->rcode) != NOERROR) 
	|| (ntohs(result_header->ancount) <= 0))
    {
	return;
    }

    // Parse out the Question section and skip over it to get the the
    // RR's following it
    // (see RFC 1035-4.1.2)
    char host[1024];
    int name_len = dn_expand(dns_result,
			     dns_result + dns_result_sz,
			     dns_result + sizeof(HEADER),
			     host, sizeof(host));
    if (name_len < 0)
    {
	return;
    }

    // The first RR is found immediately following the QUESTION
    // section, so calculate the pointer to the first RR by summing up
    // what we know about the result:
    //    1.) Pointer to result + 
    //    2.) sizeof HEADER + 
    //    3.) expanded question name (QNAME) +
    //    4.) QTYPE and QCLASS; two octets a piece
    unsigned char* rrptr = dns_result + sizeof(HEADER) + name_len + 4;


    // Create a cache for A records
    map<string, string> dns_a_record_cache;
    map<int, SRVRec>    dns_srv_record_cache;

    // Walk the RR's, building a list of targets
    while (rrptr < (dns_result + dns_result_sz))
    {

	// Expand the domain name of the current RR (RFC1034 - 4.1.3)
	name_len = dn_expand(dns_result, dns_result + dns_result_sz, rrptr,
			     host, sizeof(host));
	if (name_len < 0)
	{
	    return;
	}

	// Jump to the RR info (skipping over the NAME field)
	rrptr += name_len;
	int rrtype      = (rrptr[0] << 8 | rrptr[1]); // RRCLASS
	int rrpayloadsz = (rrptr[8] << 8 | rrptr[9]); // RDLENGTH
	rrptr += 10; // Skip over RRTYPE, RRCLASS, RRTTL, RDLENGTH

	// Process the RR
	switch (rrtype)
	{
	case T_A:
	    // A records should be saved for the duration of the lookup
	    dns_a_record_cache.insert(make_pair(string(host), Address::inet_ntoa(rrptr)));
	    break;
	case T_SRV:
	    // Expand the target name
	    name_len = dn_expand(dns_result, dns_result + dns_result_sz,
				 rrptr + 6,
				 host, sizeof(host));
	    if (name_len < 0)
	    {
		return;
	    }

	    // Cache the SRV record
	    SRVRec r = { rrptr[0] << 8 | rrptr[1], // Priority
			 rrptr[4] << 8 | rrptr[5], // Port
			 host };
	    dns_srv_record_cache.insert(make_pair(r.priority, r));
	    break;
	} 

	// Increment to next RR
	rrptr += rrpayloadsz;
    }

    // Now, walk the SRV record cache and resolve each entry in the
    // list into an IP address. Hopefully these where sent along with
    // the SRV records and are now residing in our dns_a_record_cache.
    for(map<int, SRVRec>::iterator it = dns_srv_record_cache.begin();
	it != dns_srv_record_cache.end();
	it++)
    {
	// Check A record cache..
	map<string, string>::iterator arec_it = dns_a_record_cache.find(it->second.host);
	if (arec_it != dns_a_record_cache.end())
	{
	    result->addEntry(it->second.priority, arec_it->second, it->second.port);
	}
    }
}




