//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      Socket.cpp
// Description:   bedrock::net::Socket implementation
// Created at:    Thu Jul 19 10:00:51 2001
// Modified at:   Mon Jul 30 19:29:28 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
// $Id: Socket.cpp,v 1.1 2001/12/16 20:02:26 dr_pepper Exp $
//============================================================================
#include "bedrock.h"
using namespace bedrock;
using namespace bedrock::net;
using namespace std;

Socket::Socket(SocketWatcher& watcher, ThreadPool& pool, int key, int socketfd,
	       Karma* karma, ConnectionLimit* conn_limit)
    : _threadpool(pool), _threadkey(key), _socket(socketfd), _listener(NULL),
      _current_write_buffer(NULL), _state(Socket::Created), _resolving(false),
      _conn_limit(conn_limit), _watcher(watcher)
{
    _karma = karma ? new Karma(*karma, _threadpool, _threadkey) : 0;
}

Socket::~Socket()
{
    assert(!_resolving);
    if (_karma)
	delete _karma;
}

void Socket::connect(SocketWatcher::SocketInfoHandle handle, const char* hostname, 
		     const char* service, int defaultport)
{
    // Create the underlying O/S socket
    _socket = socket(AF_INET, SOCK_STREAM, 0);
    if (_socket < 0)
        throw exception::ErrorCreatingSocket(errno, strerror(errno));

    // Make sure the socket is non-blocking
    int flags = fcntl(_socket, F_GETFL, 0);
    flags |= O_NONBLOCK;
    fcntl(_socket, F_SETFL, flags);

    // Setup Address struct
    _address.setHostname(hostname);
    _address.setService(service);
    _address.setDefaultPort(defaultport);

    _state = Socket::Connecting;
    _watcher_info = handle;

    // Queue up resolver callback
    _resolving = true;
    _watcher.getResolver().lookup(&_address, 
				  callback::wrap(this, &Socket::onConnectResolved), 
				  &_threadpool, 
				  _threadkey);
}

void Socket::onConnectResolved()
{
    // This is called by the Resolver when we have asked it to lookup
    // a hostname.  It should have filled in our _address with the
    // result (or left the _ip blank on error).  We are now ready
    // to try to connect.
    _resolving = false;

    // If we are closed, then we were just hanging around to get
    // this response.
    if (_state == Socket::Closed)
    {
        _threadpool.enqueue(
            callback::wrap(&_watcher, &SocketWatcher::cleanupSocket, _watcher_info),
            _threadkey);
        return;
    }

    // If address resolution failed, we need to notify the
    // listener that an error occurred
    if (_address.getIP().empty())
    {
        executeError(exception::InvalidHostname, 0);
        return;
    }

    // Otherwise, we've got all the necessary info to connect -- let's 
    // get started
    struct sockaddr_in addr;
    
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port   = htons(_address.getPort());

    // this next one is inet_addr except that it returns an error code
    // (0 if the addr is invalid)  
    addr.sin_addr.s_addr = inet_addr(_address.getIP().c_str());

    if (_listener)
    {
	_listener->onInitialize(this);
    }

    // Attempt to connect
    if (::connect(_socket, (struct sockaddr*)&addr, sizeof(addr)) < 0)
    {
        // Check to see if this connect just didn't complete immediately
        if (errno == EINPROGRESS)
        {
            _state = Socket::Connecting;
            _watcher.watchSocket(_watcher_info, SocketWatcher::Write);
        }
        // Something more disastrous happened; let's notify the listener
        else
        {
            executeError(exception::SocketConnectError, errno);
            return;
        }
    }
    else
    {
        // Connection is good to go; lets notify the Watcher to keep us
        // informed of reads and let the listener know that the connection
        // is ready
        executeConnect();
    }
}

void Socket::executeConnect(void)
{
    _state = Socket::Connected;

    if (!_writequeue.empty())
        executeWrite();

    if (_listener != NULL)
        _listener->onConnected(this);
}

void Socket::accept(SocketWatcher::SocketInfoHandle handle, const char* ip, 
		    unsigned int port, bool reuseaddr)
{
    // Create the underlying O/S socket
    _socket = socket(AF_INET, SOCK_STREAM, 0);
    if (_socket < 0)
        throw exception::ErrorCreatingSocket(errno, strerror(errno));

    if (reuseaddr)
    {
        int flag = 1;
        setsockopt(_socket, SOL_SOCKET, SO_REUSEADDR, (char*)&flag, sizeof(flag));
    }

    // Make sure the socket is non-blocking
    int flags = fcntl(_socket, F_GETFL, 0);
    flags |= O_NONBLOCK;
    fcntl(_socket, F_SETFL, flags);

    // Setup socket address
    sockaddr_in servaddr;
    memset((void *)&servaddr, 0, sizeof(sockaddr_in));
    servaddr.sin_family = AF_INET;
    servaddr.sin_port   = htons(port);
    if (ip == NULL || *ip == '\0')
    {
        servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    }
    else
    {
	servaddr.sin_addr.s_addr = inet_addr(ip);
    }

    if (bind(_socket, (sockaddr*)&servaddr, sizeof(sockaddr_in)) != 0)
    {
	::close(_socket);
        throw exception::BindError(errno, strerror(errno));
    }

    if (_listener)
    {
	_listener->onInitialize(this);
    }

    // Start listening on the socket
    if (listen(_socket, 5) != 0)
    {
	::close(_socket);
        throw exception::ListenError(errno, strerror(errno));
    }

    // Set our state:
    _state = Socket::Listening;

    _watcher_info = handle;

    _watcher.watchSocket(_watcher_info, SocketWatcher::Read);
}

/**
 *  Inform the socket watcher that this socket is ready to
 *  accept more connections.
 */
void Socket::resumeAccept()
{
    assert(_state == Socket::Listening);
    _threadpool.enqueue(
        callback::wrap(this, &Socket::requestAccept),
        _threadkey);
}

void Socket::requestAccept(void)
{
    if ((_state != Socket::Closed &&
         _state != Socket::Error))
        _watcher.watchSocket(_watcher_info, SocketWatcher::Read);
}

void Socket::executeAccept(void)
{
    Socket* newSocket;
    int newfd, newThreadID;
    bool keep_going = true;

    while (keep_going && (newfd = ::accept(_socket, NULL, 0)) >= 0)
    {
        struct sockaddr_in name;
#ifndef SOLARIS_2_6	
	socklen_t namelen;
#else
	int namelen;
#endif
        string addr;

        namelen = sizeof(sockaddr_in);
        if (getpeername(newfd, (sockaddr*)&name, &namelen))
        {
            ::close(newfd);
            executeError(exception::InvalidSocketAccept, errno);
            return;
        }

	// Should we accept it?
	// XXX:  I question the portability of treating an ip as an unsigned
	if (_conn_limit && !_conn_limit->allowConnection(name.sin_addr.s_addr))
	{
	    ::close(newfd);
	    continue;
	}

        // Make sure this new socket is non blocking
        int flags = fcntl(newfd, F_GETFL, 0);
        flags |= O_NONBLOCK;
        fcntl(newfd, F_SETFL, flags);

        newThreadID = _threadpool.getNextThreadID();
        newSocket = new Socket(_watcher, _threadpool, newThreadID, newfd,
                               _karma);
	newSocket->_address.addEntry(0, Address::inet_ntoa((unsigned char *)&name.sin_addr), 
				     ntohs(name.sin_port));
        try 
	{
            newSocket->_watcher_info = _watcher.acquireHandle(newSocket);
            newSocket->_state   = Socket::Connected;
            if (_listener != NULL)
		keep_going = _listener->onAccept(newSocket);

	    // If the newSocket now has a listener, queue up an onConnected event
	    if (newSocket->_listener != NULL)
		_threadpool.enqueue(callback::wrap(newSocket, &Socket::executeConnect),
				    newThreadID);
        } 
	catch (SocketWatcher::exception::WatcherFull& e) 
	{
            newSocket->_state = Socket::Closed;
            ::close(newSocket->_socket);
            delete newSocket;
            executeError(exception::SocketWatcherFull, 0);
        }
    }
    if (keep_going)
    {
        if (errno == EAGAIN || errno == EWOULDBLOCK)
            _watcher.watchSocket(_watcher_info, SocketWatcher::Read);
        else
            executeError(exception::SocketAcceptError, errno);
    }
}


/**
 *  Inform the socket watcher that this socket is ready
 *  for more data.
 */
void Socket::resumeRead()
{
    _threadpool.enqueue(
        callback::wrap(this, &Socket::requestRead),
        _threadkey);
}

void Socket::requestRead(void)
{
    assert(_state != Socket::Listening);
    if (_state != Socket::Closed &&
        _state != Socket::Error)
        _watcher.watchSocket(_watcher_info, SocketWatcher::Read);
}

void Socket::executeRead(void)
{
    int  bufcnt = 1; // Prime the pump just to get into the loop
    bool keep_going = true;
    char read_buf[4096];
    bool karma_still_good = (_karma ? _karma->mayReadMore() : true);

    while (keep_going && karma_still_good)
    {
	if(keep_going)
	{
	    bufcnt = recv(_socket, read_buf, sizeof(read_buf)-1, 0);
	}

	if (bufcnt <= 0)
	{
	    break;
	}

	if (_karma)
	{
	    karma_still_good = _karma->mayReadMore(bufcnt);
	}

        // Null-terminate and notify the listener
        read_buf[bufcnt] = '\0';
        if (_listener != NULL)
            keep_going = _listener->onRead(this, read_buf, bufcnt);
    }

    if (keep_going)
    {
        if (bufcnt == 0) 
        {
	    close();
        }
        else if (errno == EAGAIN)
	{
	    if (karma_still_good)
	    {
		_watcher.watchSocket(_watcher_info, SocketWatcher::Read);
	    }
	    else 
	    {
		_karma->callWhenReset(callback::wrap(this, &Socket::requestRead),
				      Karma::Read);
	    }
        }
        else 
        {
            executeError(exception::SocketRecvError, errno);
        }
    }
}


/**
 *  Write data on the socket.  If the socket is not yet connected
 *  it will queue it for writing when the socket connection is
 *  completed.  Do not call this from an accepting socket or you
 *  will get an error.
 */
void Socket::write(const char* data)
{
    _threadpool.enqueue(
        callback::wrap(this, &Socket::requestWrite, strdup(data)),
        _threadkey);
}

void Socket::requestWrite(char *data)
{
    assert(_state != Socket::Listening);
    if (_state != Socket::Closed &&
        _state != Socket::Error)
    {
        _writequeue.push_back(data);

        if (_state == Socket::Connected)
            executeWrite();
    }
    else
    {
        // TODO: Bounce the data instead of just dumping it??
        cerr << "Dropping: " << data << endl;
        free(data);
    }
}

void Socket::executeWrite(void)
{
    int cnt;
    bool karma_still_good = (_karma ? _karma->mayWriteMore() : true);

    while (!_writequeue.empty())
    {
	if (!karma_still_good)
	{
	    _karma->callWhenReset(callback::wrap(this, &Socket::executeWrite),
				  Karma::Write);
	    break;
	}

        if (_current_write_buffer == NULL)
        {
            _current_write_buffer = _writequeue.front();
            _current_write_buffer_len = strlen(_current_write_buffer);
        }

        cnt = send(_socket, _current_write_buffer, _current_write_buffer_len, 0);
	if (cnt > 0 && _karma)
	{
	    karma_still_good = _karma->mayWriteMore(cnt);
	}

        if (cnt == _current_write_buffer_len)
        {
            // This buffer was allocated in Socket::write by a call
            // to strdup(), so we have to call free as opposed to
            // delete'ing it.
            free(_writequeue.front());
            _writequeue.pop_front();
            _current_write_buffer = NULL;
        }
        else if (cnt < 0)
        {
            if (errno == EAGAIN)
            {
                _watcher.watchSocket(_watcher_info, SocketWatcher::Write);
            }
            else
            {
                executeError(exception::SocketSendError, errno);
            }
            break;
        }
        else 
        {
            _current_write_buffer += cnt;
            _current_write_buffer_len -= cnt;
        }
    }
}

void Socket::onSocketEvent(short int revents)
{
    // First of all, if there was an error...
    if (revents & (POLLERR | POLLNVAL | POLLHUP))
    {
        // Per Stevens:171, calling a read after a poll error
        // will set errno for us so that we know what happened;
        // at least, that's the idea :) getsockopt doesn't seem 
        // to work very well in threaded environments.
        char buf;
        read(_socket, &buf, sizeof(buf));

        executeError(exception::SocketErrorEvent, errno);
        return;
    }

    // Well, we didn't get an error, so...
    
    // If we are a listening socket, then we are only looking for read events,
    // as that is what we receive when a new socket is ready to be accepted.
    if (_state == Socket::Listening)
    {
        if (revents & POLLIN)
        {
            executeAccept();
        }
    }
    // Otherwise, if we are a connected socket, then we are looking for read
    // and/or write events.
    else if (_state == Socket::Connected)
    {
        if (revents & POLLIN)
        {
            executeRead();
        }

        if (revents & POLLOUT)
        {
            executeWrite();
        }
    }
    // And finally, if we are in the process of connecting, then we want to get
    // a write event without a read event.
    else if (_state == Socket::Connecting)
    {
        struct sockaddr_in name;
#ifndef SOLARIS_2_6	
	socklen_t namelen;
#else
	int namelen;
#endif

        namelen = sizeof(sockaddr_in);
        if (getpeername(_socket, (sockaddr*)&name, &namelen))
            executeError(exception::SocketConnectError, errno);
        else
            executeConnect();
    }
}

/**
 *  Close the socket.  This also eventually destroys the socket
 *  object, use as your destructor.
 */
void Socket::close()
{
    _watcher.watchSocket(_watcher_info, SocketWatcher::Closing);
}

void Socket::executeClose(void)
{
    // And finally close the socket:
    ::close(_socket);
    _socket = -1;

    _state = Socket::Closed;

    // Tell our listener that we are closed:
    if (_listener != NULL)
    {
        _listener->onClose(this);
        _listener = NULL;
    }


    if (!_resolving)
    {
        _threadpool.enqueue(callback::wrap(&_watcher, 
					   &SocketWatcher::cleanupSocket, 
					   _watcher_info),
			    _threadkey);
    }
}

void Socket::executeError(exception::ErrorCode ec, int err)
{
    // Tell our listener that there was an error:
    _state = Socket::Error;
    if (_listener != NULL)
        _listener->onError(this, ec, err);
}

