//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      SocketWatcher.cpp
// Description:   bedrock::net::SocketWatcher implementation
// Created at:    Thu Jul 19 10:00:51 2001
// Modified at:   Wed Aug  1 21:23:28 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
// $Id: SocketWatcher.cpp,v 1.1 2001/12/16 20:02:26 dr_pepper Exp $
//============================================================================
#include "bedrock.h"
using namespace bedrock;
using namespace bedrock::net;
using namespace std;


/**
 *  Default constructor.
 *  @param pool Thread pool for the watcher to queue events on.
 *  @param maxsockets Maximum number of sockets that this watcher 
 *    will need to deal with. 
 */
SocketWatcher::SocketWatcher(ThreadPool& pool, unsigned int maxsockets)
    : _resolver(3), _socket_cnt(maxsockets+1), _threadpool(pool), 
      _close_pending_count(0), _state(SocketWatcher::Running), _stopped_cv(_lock)
{
    // Initialize _available_slots; yes, we start at 1 since 0 is reserved.
    for (unsigned int i = 1; i < _socket_cnt; i++)
        _available_slots.insert(i);
    
    _fds.reserve(_socket_cnt);

    for (unsigned int i = 0; i < _socket_cnt; i++)
    {
	pollfd fd = { -1, 0, 0 };
	_fds[i] = fd;
    }

    _watched_sockets.reserve(_socket_cnt);

    // Hack for Linux; poll() is messed up.
    dup2(0, _socket_cnt + 3);

    // Initialize our wakeup pipes
    int pipefds[2];    
    pipe(pipefds);

    // Init the read pipe -- we need to make sure it's non-blocking
    _fds[0].fd        = pipefds[0];
    _fds[0].events    = POLLIN;

    int flags = fcntl(_fds[0].fd, F_GETFL, 0);
    flags |= O_NONBLOCK;
    fcntl(_fds[0].fd, F_SETFL, flags);

    // Init the write pipe
    _poll_wakeup_pipe = pipefds[1];

    // Start the watcher thread
    pthread_create(&_watcher_thread, 
                   NULL,
                   SocketWatcher::beginProcessing, 
                   this);
}

SocketWatcher::~SocketWatcher()
{
    // Stop the watcher thread and set the state appropriately
    {
	MutexProxy mp(_lock);

	_state = SocketWatcher::Shutdown;

	pthread_cancel(_watcher_thread);

	char buf = 'w';
	write(_poll_wakeup_pipe, &buf, sizeof(char));
    }

    // Wait for the watcher's thread to finish
    pthread_join(_watcher_thread, NULL);

    MutexProxy mp(_lock);
    // Start the cleanup process of all known sockets
    {
	// Clear the _unwatched_sockets list
	_unwatched_sockets.clear();

	// Notify all the sockets we know about that the manager is closing shop
	for_each(_sockets.begin(), _sockets.end(), _closeSocket);
    }

    // Close the wakeup pipe
    close(_fds[0].fd);
    close(_poll_wakeup_pipe);
    
    // Wait for all the sockets to complete their cleanup
    if (!_sockets.empty())
    {
        while (_state != SocketWatcher::Stopped)
            _stopped_cv.wait();
    }

}

void SocketWatcher::_closeSocket(SocketInfo& si)
{
    if (si.event != SocketWatcher::Closed)
    {
	si.socket->processClose();
    }
}

/**
 *  Instantiate a new socket to accept incoming connections.
 *  @param listener Event handler object to be associated with the new Socket
 *  @param ip IP to listen on for incoming connections
 *  @param port Port to listen on for incoming connections
 *  @returns a reference to a new Socket
 */
Socket* SocketWatcher::createListenSocket(SocketEventListener* listener,
                                          const char* ip, 
                                          int port,
                                          bool reuseaddr,
                                          int thread_key,
					  Karma *karma,
					  ConnectionLimit* conn_limit)
{
    // Create the socket:
    Socket* result = new Socket(*this, _threadpool, 
				thread_key == -1 ? _threadpool.getNextThreadID() : thread_key, -1, karma, conn_limit);
    // Set the socket's listener:
    result->setListener(listener);

    // Request a slot for the new socket and start listening for
    // incoming connections
    SocketInfoHandle handle;
    try
    {
	handle = acquireHandle(result);
	result->accept(handle, ip, port, reuseaddr);
    }
    // For any other Socket::exception, we need to be sure to release
    // the allocated socket _and_ release the handle we got from the SocketWatcher
    catch (Socket::exception::BaseException& e)
    {
	releaseHandle(handle);
	delete result;
	throw;
    }
    // Any other exception, just release the allocated Socket
    catch (...)
    {
	delete result;
	throw;
    }

    return result;
}

/**
 *  Instantiate a new socket that will open an outbound TCP 
 *  connection to another host.
 *  @param listener Event handler object to be associated with the new Socket
 *  @param hostname DNS hostname to open a connection to
 *  @param service  SRV service identifier (optional)
 *  @param defaultport Port number to connect to if no service is resolved
 *  @returns a reference to a new Socket
 */
Socket* SocketWatcher::createConnectSocket(SocketEventListener* listener,
                                           const char* host, 
                                           const char* service, 
                                           int defaultport,
                                           int thread_key,
					   Karma *karma)
{
    // Create the socket:
    Socket* result = new Socket(*this, _threadpool, 
				thread_key == -1 ? _threadpool.getNextThreadID() : thread_key, -1, karma, 0);

    // Set the socket's listener:
    result->setListener(listener);

    // Request a slot for this new socket and fire off the connection sequence
    SocketInfoHandle handle;
    try
    {
	handle = acquireHandle(result);	
	result->connect(handle, host, service, defaultport);
    }
    // For any other Socket::exception, we need to be sure to release
    // the allocated socket _and_ release the handle we got from the SocketWatcher
    catch (Socket::exception::BaseException& e)
    {
	releaseHandle(handle);
	delete result;
	throw;
    }
    // Any other exception, just release the allocated Socket
    catch (...)
    {
	delete result;
	throw;
    }

    return result;
}

/**
 *  Put this socket in our internal structures.
 */
SocketWatcher::SocketInfoHandle SocketWatcher::acquireHandle(Socket* s)
{
    SocketInfoHandle result;

    // Create a new SocketInfoHandle
    SocketInfo si = { -1, s, SocketWatcher::NOOP };

    MutexProxy mp(_lock);
   
    // A free slot was available..
    if (!_available_slots.empty())
    {
	// Ok, let's get the next available slot id:
	set<int>::iterator it = _available_slots.begin(); 	
	si.id = *it;
	_available_slots.erase(it);
	// Register this SocketInfo with our global list
	result = _sockets.insert(_sockets.end(), si);
	// Stick a handle to this SocketInfo into the watched sockets vector
	_watched_sockets[result->id] = result;
    }
    // There were no free slots available.
    else
    {
	// If the number of closing sockets is less than the number of
	// sockets waiting on those others to close, go ahead and
	// queue up this socket to be watched whenever possible
	if (_close_pending_count > (int)_unwatched_sockets.size())
	{
	    result = _sockets.insert(_sockets.end(), si);
	    _unwatched_sockets.push_back(result);
	}
	// Otherwise throw an exception indicating that we were unable
	// to watch a new socket due to lack of resources
	else
	{
	    throw exception::WatcherFull();
	}
    }

    return result;
}

/**
   Emergency release for info handle; only called from exception
   handlers
*/
void SocketWatcher::releaseHandle(SocketInfoHandle& handle)
{
    MutexProxy mp(_lock);
    if (handle->id > 0)
    {
	_available_slots.insert(handle->id);
    }
    else
    {
	_unwatched_sockets.remove(handle);
    }
    _sockets.erase(handle);
}

/**
 *  Update the events a socket is being watched for
 */
/**
 *  Update the events a socket is being watched for
 */
void SocketWatcher::watchSocket(const SocketInfoHandle& handle, SocketWatcher::Event event)
{
    MutexProxy   mp(_lock);
    bool         change_made = false;

    if ((handle->event == SocketWatcher::Closing) ||
	(handle->event == SocketWatcher::Closed))
    {
	return;
    }

    bool pending_empty = _pending.empty();

    // Shortcut; if the current event is a NOOP, any other event overrides
    // and forces a change..
    if ((handle->event == NOOP) && (event != NOOP))
    {
	handle->event = event;
	change_made = true;

	if (handle->event == SocketWatcher::Closing)
	    _close_pending_count++;

	if (handle->id > 0)
	    _pending.push_back(handle->id);
    }
    // Otherwise, we have to run through a more sophisticated
    // algorithm for dealing the different states
    else if (handle->event != event)
    {
	switch(event)
	{
	case SocketWatcher::Closing:
	    _close_pending_count++;
	    handle->event = event;
	    change_made   = true;
	    break;
	case SocketWatcher::Write:
	    if (handle->event == SocketWatcher::Read)
	    {
		handle->event = SocketWatcher::ReadWrite;
		change_made   = true;
	    }
	    break;
	case SocketWatcher::Read:
	    if (handle->event == SocketWatcher::Write)
	    {
		handle->event = SocketWatcher::ReadWrite;
		change_made = true;
	    }
	    break;
	case SocketWatcher::ReadWrite:
            handle->event = SocketWatcher::ReadWrite;
            change_made = true;
	case SocketWatcher::Closed:
	case SocketWatcher::NOOP:
	    break;
	}
    }

    if (change_made && pending_empty && handle->id > 0)
    {
	char buf = 'w';
	write(_poll_wakeup_pipe, &buf, sizeof(char));
    }
}

/**
 *  Function called exclusively by Socket::executeClose to let
 *  the SocketWatcher know when it is ok to delete the Socket object
 */
void SocketWatcher::cleanupSocket(SocketInfoHandle handle)
{
    MutexProxy mp(_lock);

    // Check to see if there are any unwatched sockets waiting on a
    // slot to become available. If so, reuse the closing Sockets info
    // for one of the unwatched sockets
    if (!_unwatched_sockets.empty())
    {
	// Grab the handle from the unwatched sockets list and set
	// it up as a watched socket. 
	SocketInfoHandle h = _unwatched_sockets.front();
	_unwatched_sockets.pop_front();

	h->id                   = handle->id;
	_watched_sockets[h->id] = h;

	// If there is an event already requested on this socket,
	// make sure it's registered in the pending list
	if (h->event != NOOP)
	{
	    if (_pending.empty())
	    {
		char buf = 'w';
		write(_poll_wakeup_pipe, &buf, sizeof(char));
	    }
	    _pending.push_back(h->id);
	}
    }
    // Otherwise, release all the socket info
    else
    {
	_available_slots.insert(handle->id);
    }
    // Always delete the actual Socket object
    delete handle->socket;
    // Remove the info on this socket from our global
    // list of socket info
    _sockets.erase(handle);
    
    if (_state != SocketWatcher::Shutdown)
    {
	// We only decrement the close pending counter when our
	// SocketWatcher is still functional. Otherwise, the counter
	// would drop below zero when the SocketWatcher was shutting
	// down since we shortcut Socket close procedures to force
	// all sockets to close
	_close_pending_count--;
    }
    else
    {
	if (_sockets.empty())
	{
	    _state = SocketWatcher::Stopped;
	    _stopped_cv.signal();
	}
    }
}

/**
 *  Static thread entry point.
 */
void* SocketWatcher::beginProcessing(void* arg)
{
    SocketWatcher* watcher = (SocketWatcher *)arg;
    watcher->process();
    return NULL;
}

void _cancellation_cleanup(void* arg)
{
    Mutex* m = (Mutex*)arg;
    m->unlock();
}

/**
 *  The main poll/process loop.
 */
void SocketWatcher::process(void)
{
    while (1)
    {
	int rc = poll(&_fds[0], _socket_cnt, -1);
	if (rc < 0 && errno != EINTR)
        {
	    string error = "Poll Error: ";
	    error += strerror(errno);
	    Application::exit(1, error);
        }

        pthread_testcancel();

	pthread_cleanup_push(_cancellation_cleanup, &_lock);
	_lock.lock();

	if (!_sockets.empty())
	{
	    processPending();
	    processEvents();
	}

	_lock.unlock();
	pthread_cleanup_pop(0);
    }
}

/**
 *  Process the poll events.  Called from process.
 */
void SocketWatcher::processEvents(void)
{
    for (unsigned int i = 1; i < _socket_cnt; i++)
    {
        if (_fds[i].revents)
        {
            // This just puts a callback on the socket's queue,
            // so it is safe to call while locked
	    _watched_sockets[i]->socket->processEvents(_fds[i].revents);

            if (_fds[i].revents & (POLLERR|POLLNVAL|POLLHUP))
                _fds[i].fd = -1;

            // Reset the events for which we are looking.
            _fds[i].events &= ~_fds[i].revents;
            _fds[i].revents = 0;
        }
    }
}

/**
 *  Process the pending events, modifying the pollfd structs
 *  as necessary.  Called from process.
 */
void SocketWatcher::processPending(void)
{
    for (list<int>::iterator it = _pending.begin();
         it != _pending.end();
         it++)
    {
        int id = *it;

        _fds[id].fd = _watched_sockets[id]->socket->getFD();

        switch (_watched_sockets[id]->event)
        {
        case SocketWatcher::NOOP:
            break;
        case SocketWatcher::Read:
            _fds[id].events |= POLLIN;
            break;
        case SocketWatcher::Write:
            _fds[id].events |= POLLOUT;
            break;
        case SocketWatcher::ReadWrite:
            _fds[id].events |= (POLLIN | POLLOUT);
            break;
        case SocketWatcher::Closing:
            {
                _fds[id].fd      = -1;
                _fds[id].events  =  0;
                _fds[id].revents =  0;
                // Tell the socket to cleanup
		_watched_sockets[id]->event = SocketWatcher::Closed;
                _watched_sockets[id]->socket->processClose();
            }
            break;
	default:
	    break;
        }
	if (_watched_sockets[id]->event != SocketWatcher::Closed)
	{
	    _watched_sockets[id]->event = NOOP;
	}
    }

    char buf[100]; 
    while ( read(_fds[0].fd, &buf, sizeof(buf)) != -1); 

    _pending.clear();

}
